"""
Name: Klein Radiation Perfect Fluid
Coordinates: Spherical
Symmetry:
    - Spherical
    - Static
"""
from sympy import Rational, diag, sin, sqrt, symbols

coords = symbols("t r theta phi", real=True)
variables = symbols("p_0", constant=True)
functions = ()
t, r, th, ph = coords
p0 = variables
metric = diag(-sqrt(7 * p0) * r, Rational(7, 4), r ** 2, r ** 2 * sin(th) ** 2)



"""
Name: Durgapal and Fuloria Perfect Fluid
References: Durgapal et al., Gen. Rel. Grav., v17, p671-681, (1985)
Coordinates: Spherical
Symmetry:
    - Spherical
    - Static
"""
from sympy import diag, sin, symbols

coords = symbols("t r theta phi", real=True)
variables = symbols("B C", constant=True)
functions = ()
t, r, th, ph = coords
B, C = variables
expr = 1 + C * r ** 2
metric = diag(
    -B * expr ** 4,
    7 * expr ** 2 / (-(expr ** 2) - 8 * expr + 16),
    r ** 2,
    r ** 2 * sin(th) ** 2,
)



"""
Name: Griffiths
References: Griffiths, Math. Proc. Camb. Phil. Soc., v77, p559, (1975)
"""
from sympy import Rational, symbols, zeros

coords = symbols("u v x y", real=True)
variables = symbols("a", constant=True)
functions = ()
u, v, x, y = coords
a = variables
metric = zeros(4)
metric[2, 2] = (
    Rational(3, 2)
    + 4 * a ** 2 * v * u
    + 6 * a ** 2 * v ** 2
    + 6 * a ** 2 * u ** 2
    + 4 * a ** 2 * y * u
    + 2 * a ** 2 * y ** 2
    + 4 * a ** 2 * v * y
)
metric[3, 3] = -Rational(1, 2)
metric[0, 1] = metric[1, 0] = -1
metric[0, 2] = metric[2, 0] = -a * (2 * v - y)
metric[0, 3] = metric[3, 0] = -1
metric[1, 2] = metric[2, 1] = -a * (2 * u - y)
metric[1, 3] = metric[3, 1] = -1
metric[2, 3] = metric[3, 2] = a * (v + u + 2 * y)



"""
Name: Beckers, Sinzinkayo, and Demaret
References: Beckers et al., Phys. Rev. D, v30, p1846, (1984)
Coordinates: Cartesian
Notes:
    - k = 1
    - d = 0
"""
from sympy import diag, symbols

coords = symbols("t x y z", real=True)
variables = symbols("m", constant=True)
functions = ()
t, x, y, z = coords
m = variables
expr = x ** (2 * m)
metric = diag(-expr, expr, expr, expr)



"""
Name: Pant and Sah
References: Pant et al., J. Math. Phys., v20, p2537-2539, (1979)
Coordinates: Spherical
Symmetry:
    - Spherical
    - Static
"""
from sympy import diag, sin, symbols

coords = symbols("t r theta phi", real=True)
variables = symbols("A n", constant=True)
functions = ()
t, r, th, ph = coords
A, n = variables
metric = diag(-A * r ** (2 * n), n ** 2 + 1, r ** 2, r ** 2 * sin(th) ** 2)



"""
Name: Heintzmann Perfect Fluid
References: Heintzmann, Z. Phys., v228, p489-493, (1969)
Coordinates: Spherical
Symmetry:
    - Spherical
    - Static
"""
from sympy import diag, sin, symbols

coords = symbols("t r theta phi", real=True)
variables = symbols("A a K", constant=True)
functions = ()
t, r, th, ph = coords
A, a, K = variables
metric = diag(
    -(A ** 2) * (1 + a * r ** 2) ** 3,
    (1 + a * r ** 2) / K,
    r ** 2,
    r ** 2 * sin(th) ** 2,
)



"""
Name: Einstein
References: Stephani (10.23a) p122
Coordinates: Spherical
Symmetry: Static
"""
from sympy import diag, sin, symbols

coords = symbols("t r theta phi", real=True)
variables = symbols("Lambda", constant=True)
functions = ()
t, r, th, ph = coords
La = variables
metric = diag(-1, 1 / (1 - La * r ** 2), r ** 2, r ** 2 * sin(th) ** 2)



"""
Name: Bertotti-Robinson
References:
    - Bertotti, Phys. Rev., v116, p1331, (1959)
    - Lovelock, Commun. Math. Phys., v5, p257, (1967)
    - Dolan, Commun. Math. Phys., v9, p161, (1968)
    - Stephani (10.18) p121
Coordinates: Spherical
"""
from sympy import diag, sin, symbols

coords = symbols("t r theta phi", real=True)
variables = symbols("q", constant=True)
functions = ()
t, r, th, ph = coords
q = variables
metric = diag(-(q ** 2) / r ** 2, q ** 2 / r ** 2, q ** 2, q ** 2 * sin(th) ** 2)



"""
Name: Tariq and Tupper
References:
    - Tariq et al., Gen. Rel. Grav., v6, p345, (1975)
    - Stephani (11.64) p138
Coordinates: Cartesian
Notes: Admits G3VIo on S3
"""
from sympy import diag, exp, symbols

coords = symbols("t x y z", real=True)
variables = ()
functions = ()
t, x, y, z = coords
metric = diag(-1, 4 * t ** 2 / 3, t / exp(2 * x), t * exp(2 * x))



"""
Name: Flat Boost Isotropy
References: Stephani (11.16) p128
Symmetry: Boost Rotation
Notes: Temporal Sine
"""
from sympy import Function, diag, sin, symbols

coords = symbols("t w x y", real=True)
variables = ()
functions = symbols("alpha beta", cls=Function)
t, w, x, y = coords
al, be = functions
metric = diag(-be(w) ** 2 * sin(y) ** 2, 1, al(w) ** 2, be(w) ** 2)



from riccipy.metric import *
from riccipy.partial import *
from riccipy.tensor import *
from sympy import Expr, diag, eye, sin, symbols, tensorproduct, zeros


def _generate_schwarzschild():
    coords = symbols("t r theta phi", real=True)
    t, r, th, ph = coords
    schw = diag(1 - 1 / r, -1 / (1 - 1 / r), -(r ** 2), -(r ** 2) * sin(th) ** 2)
    g = SpacetimeMetric("g", coords, schw, timelike=True)
    mu, nu = indices("mu nu", g)
    return (coords, t, r, th, ph, schw, g, mu, nu)


def test_Metric():
    (coords, t, r, th, ph, schw, g, mu, nu) = _generate_schwarzschild()
    assert isinstance(g, AbstractTensor)
    assert isinstance(g, TensorIndexType)
    assert isinstance(g(mu, nu), IndexedTensor)
    assert isinstance(g.metric, Tensor)
    assert isinstance(g.partial, PartialDerivative)
    assert g.metric.covar == (-1, -1)


def test_SpacetimeMetric():
    (coords, t, r, th, ph, schw, g, mu, nu) = _generate_schwarzschild()
    array1 = g.as_array()
    assert g.signature == (1, -1, -1, -1)
    rev_sig = g.reverse_signature()
    array2 = g.as_array()
    assert rev_sig == (-1, 1, 1, 1)
    assert array1 == -1 * array2


def test_Metric_density():
    (coords, t, r, th, ph, schw, g, mu, nu) = _generate_schwarzschild()
    res1 = g.density()
    assert res1.equals(r ** 2 * abs(sin(th)))
    res2 = g.density(-1)
    assert res2.equals(1 / res1)


def test_Metric_determinant():
    (coords, t, r, th, ph, schw, g, mu, nu) = _generate_schwarzschild()
    res = g.determinant
    assert res.equals(-(r ** 4) * sin(th) ** 2)


def test_Metric_christoffel():
    (coords, t, r, th, ph, schw, g, mu, nu) = _generate_schwarzschild()
    gamma = g.christoffel.simplify()
    assert (
        str(gamma)
        == "[[[0, 1/(2*r*(r - 1)), 0, 0], [1/(2*r*(r - 1)), 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0]], [[(r - 1)/(2*r**3), 0, 0, 0], [0, -1/(2*r*(r - 1)), 0, 0], [0, 0, 1 - r, 0], [0, 0, 0, (1 - r)*sin(theta)**2]], [[0, 0, 0, 0], [0, 0, 1/r, 0], [0, 1/r, 0, 0], [0, 0, 0, -sin(2*theta)/2]], [[0, 0, 0, 0], [0, 0, 0, 1/r], [0, 0, 0, 1/tan(theta)], [0, 1/r, 1/tan(theta), 0]]]"
    )


def test_Metric_riemann():
    (coords, t, r, th, ph, schw, g, mu, nu) = _generate_schwarzschild()
    rh, si = indices("rho sigma", g)
    R = g.riemann

    expr = R(-rh, -si, -mu, -nu) + R(-si, -rh, -mu, -nu)
    assert all(expand_array(expr).applyfunc(lambda c: c.equals(0)).args[0])
    expr = R(-rh, -si, -mu, -nu) + R(-rh, -si, -nu, -mu)
    assert all(expand_array(expr).applyfunc(lambda c: c.equals(0)).args[0])
    expr = R(-rh, -si, -mu, -nu) - R(-mu, -nu, -rh, -si)
    assert all(expand_array(expr).applyfunc(lambda c: c.equals(0)).args[0])
    expr = R(-rh, -si, -mu, -nu) + R(-rh, -mu, -nu, -si) + R(-rh, -nu, -si, -mu)
    assert all(expand_array(expr).applyfunc(lambda c: c.equals(0)).args[0])


def test_Metric_ricci_tensor():
    (coords, t, r, th, ph, schw, g, mu, nu) = _generate_schwarzschild()
    R = g.ricci_tensor
    assert zeros(4).equals(R.as_array())


def test_Metric_ricci_scalar():
    (coords, t, r, th, ph, schw, g, mu, nu) = _generate_schwarzschild()
    R = g.ricci_scalar
    assert isinstance(R, Expr)
    assert R.equals(0)


def test_Metric_einstein():
    (coords, t, r, th, ph, schw, g, mu, nu) = _generate_schwarzschild()
    G = g.einstein
    assert zeros(4).equals(G.as_array())


def test_Metric_weyl():
    x, y, z = symbols("x y z", real=True)
    eta = Metric("eta", (x, y, z), eye(3))
    zero_tensor = tensorproduct(zeros(3, 3), zeros(3, 3))
    assert eta.weyl.as_array() == zero_tensor

    (coords, t, r, th, ph, schw, g, mu, nu) = _generate_schwarzschild()
    rh, si = indices("rho sigma", g)
    C = g.weyl

    expr = C(-rh, -si, -mu, -nu) + C(-si, -rh, -mu, -nu)
    assert all(expand_array(expr).applyfunc(lambda c: c.equals(0)).args[0])
    expr = C(-rh, -si, -mu, -nu) + C(-rh, -si, -nu, -mu)
    assert all(expand_array(expr).applyfunc(lambda c: c.equals(0)).args[0])
    expr = C(-rh, -si, -mu, -nu) - C(-mu, -nu, -rh, -si)
    assert all(expand_array(expr).applyfunc(lambda c: c.equals(0)).args[0])
    expr = C(-rh, -si, -mu, -nu) + C(-rh, -mu, -nu, -si) + C(-rh, -nu, -si, -mu)
    assert all(expand_array(expr).applyfunc(lambda c: c.equals(0)).args[0])



"""
Name: Levi-Civita
References: Stephani (Table 16.2) p188
Coordinates: Spherical
Symmetry: Spherical
Notes: Class A1
"""
from sympy import diag, sin, symbols

coords = symbols("t r theta phi", real=True)
variables = symbols("M", constant=True)
functions = ()
t, r, th, ph = coords
M = variables
metric = diag(-(1 - 2 * M / r), 1 / (1 - 2 * M / r), r ** 2, r ** 2 * sin(th) ** 2)



"""
Name: Godel
References:
    - Rev. Mod. Phys., v21, p447, (1949)
    - Stephani (10.25) 122
Coordinates: Cartesian
"""
from sympy import Rational, diag, exp, sqrt, symbols

coords = symbols("t x y z", real=True)
variables = symbols("omega", constant=True)
functions = ()
t, x, y, z = coords
om = variables
metric = diag(-1, 1, -Rational(1, 2) * exp(2 * sqrt(2) * om * x), 1)
metric[0, 2] = metric[2, 0] = -exp(sqrt(2) * om * x)



"""
Name: Ellis and MacCallum Vacuum
References:
    - Ellis et al., Commun. Math. Phys., v12, p108, (1969)
    - Stephani (11.56) p136
Coordinates: Cartesian
Notes: Bianchi VIo
"""
from sympy import diag, exp, sqrt, symbols

coords = symbols("t x y z", real=True)
variables = symbols("n", constant=True)
functions = ()
t, x, y, z = coords
n = variables
metric = diag(
    -sqrt(t) * exp(n ** 2 * t ** 2),
    sqrt(t) * exp(n ** 2 * t ** 2),
    t * exp(2 * n * x),
    t * exp(-2 * n * x),
)



"""
Name: Tolman Perfect Fluid
References: Tolman, Phys. Rev., v55, p363-373, (1939)
Coordinates: Spherical
Symmetry:
    - Spherical
    - Static
Notes: Type VII
"""
from sympy import Rational, diag, log, sin, sqrt, symbols

coords = symbols("t r theta phi", real=True)
variables = symbols("A B C R", constant=True)
functions = ()
t, r, th, ph = coords
A, B, C, R = variables
expr1 = 1 - r ** 2 / R ** 2 + 4 * r ** 4 / A ** 4
expr2 = sin(
    log(
        sqrt((sqrt(expr1) + 2 * r ** 2 / A ** 2 - Rational(1, 4) * A ** 2 / R ** 2) / C)
    )
)
metric = diag(-(B ** 2) * expr2 ** 2, 1 / expr1, r ** 2, r ** 2 * sin(th) ** 2)



"""
Name: Cross product of Constant Curvature Subspaces
References: Stephani (10.8) p118
Notes:
    - Temporal Hyperbolic Sine
    - Spatial Hyperbolic Sine
"""
from sympy import diag, sinh, symbols

coords = symbols("t x y z", real=True)
variables = symbols("A B", constant=True)
functions = ()
t, x, y, z = coords
A, B = variables
metric = diag(-(B ** 2) * sinh(z) ** 2, A ** 2, A ** 2 * sinh(x) ** 2, B ** 2)



"""
Name: Bianchi II
Coordinates: Cartesian
"""
from sympy import Function, diag, exp, symbols

coords = symbols("t x y z", real=True)
variables = ()
functions = symbols("alpha", cls=Function)
t, x, y, z = coords
al = functions
metric = diag(1, exp(-2 * al(t)), exp(al(t)), exp(al(t)))



"""
Name: Schwarzschild
Coordinates: Israel
Symmetry:
    - Spherical
    - Static
"""
from sympy import Rational, diag, sin, symbols

coords = symbols("u w theta phi", real=True)
variables = symbols("M", constant=True)
functions = ()
u, w, th, ph = coords
M = variables
expr = 2 * M + Rational(1, 4) * u * w / M
metric = diag(
    Rational(1, 2) * w ** 2 / (M * expr), 0, expr ** 2, expr ** 2 * sin(th) ** 2
)
metric[0, 1] = metric[1, 0] = 1



"""
Name: Szekeres Stiff Perfect Fluid
References: Szekeres, Commun. Math. Phys., v41, p55, (1975)
Coordinates: Cartesian
Notes: Abelian Coordinates
"""
from sympy import cosh, diag, exp, symbols

coords = symbols("t x y z", real=True)
variables = ()
functions = ()
t, x, y, z = coords
metric = diag(
    -exp(6 * x) / cosh(2 * t) ** 2,
    exp(6 * x) / cosh(2 * t) ** 2,
    exp(2 * x) * cosh(2 * t),
    exp(2 * x) * cosh(2 * t),
)



"""
Name: Friedman-Robertson-Walker Perfect Fluid
References:
    - Robertson, Astrophys. J., v82, p284, (1935)
    - Robertson, Astrophys. J., v83, p137, (1936)
    - Stephani (10.9) p118
Coordinates: Spherical
Notes: Flat
"""
from sympy import Function, diag, sin, symbols

coords = symbols("t r theta phi", real=True)
variables = ()
functions = symbols("a", cls=Function)
t, r, th, ph = coords
a = functions
metric = diag(-1, a(t) ** 2, a(t) ** 2 * r ** 2, a(t) ** 2 * r ** 2 * sin(th) ** 2)



"""
Name: Harrison
References:
    - Harrison, Phys. Rev., v116, p1285, (1959)
    - d'Inverno et al., J. Math. Phys., v12, p1258, (1971)
Notes:
    - Kinnersley Class II.D
    - a = l = 0
"""
from sympy import diag, exp, symbols

coords = symbols("x_0:4", real=True)
variables = ()
functions = ()
x0, x1, x2, x3 = coords
metric = diag(
    -(x3 ** 2),
    exp(4 * x2) ** 2 / (1 + x3 ** 2) ** 2,
    1 / (1 + x3 ** 2) ** 2,
    1 / (1 + x3 ** 2) ** 4,
)



"""
Name: Nariai Vacuum
References: Nariai, Sci. Rep. Tohoku Univ., v35, p62, (1951)
Coordinates: Cartesian
"""
from sympy import Function, cos, diag, log, sin, sqrt, symbols

coords = symbols("t x y z", real=True)
variables = symbols("lambda", constant=True)
functions = symbols("a b", cls=Function)
t, x, y, z = coords
la = variables
a, b = functions
expr1 = x ** 2 + y ** 2 + z ** 2
expr2 = a(t) * cos(log(sqrt(expr1) / la)) + b(t) * sin(log(sqrt(expr1) / la))
metric = diag(-expr2, la ** 2 / expr1, la ** 2 / expr1, la ** 2 / expr1)



"""
Name: Bondi
References: Bondi, Proc. Roy. Soc. Lond. A, v269, p21, (1962)
"""
from sympy import Function, exp, sin, symbols, zeros

coords = symbols("r u theta phi", real=True)
variables = ()
functions = symbols("Q U V gamma", cls=Function)
r, u, th, ph = coords
Q, U, V, ga = functions
metric = zeros(4)
metric[1, 1] = (
    -exp(2 * Q(r, u, th) + 2 * ga(r, u, th))
    * (
        V(r, u, th) * exp(-2 * ga(r, u, th))
        - exp(-2 * Q(r, u, th)) * U(r, u, th) ** 2 * r ** 3
    )
    / r
)
metric[2, 2] = r ** 2 * exp(2 * ga(r, u, th))
metric[3, 3] = r ** 2 * exp(-2 * ga(r, u, th)) * sin(th) ** 2
metric[0, 1] = metric[1, 0] = -exp(2 * Q(r, u, th))
metric[1, 2] = metric[2, 1] = -U(r, u, th) * r ** 2 * exp(2 * ga(r, u, th))



"""
Name: Taub Perfect Fluid
References:
    - Taub, Phys. Rev., v103, p454, (1956)
    - Stephani (13.44) p161
Coordinates: Cartesian
Symmetry:
    - Planar
    - Static
"""
from sympy import diag, symbols

coords = symbols("t x y z", real=True)
variables = symbols("mu", constant=True)
functions = ()
t, x, y, z = coords
mu = variables
metric = diag(-((z - 3) ** 2), z ** 2, z ** 2, 3 / (mu * z ** 2))



"""
Name: Generic Static Spherical
References: Stephani (14.1) p163
Coordinates: Spherical
Symmetry:
    - Spherical
    - Static
Notes: Additional exponential factors
"""
from sympy import Function, diag, exp, sin, symbols

coords = symbols("t r theta phi", real=True)
variables = ()
functions = symbols("alpha beta", cls=Function)
t, r, th, ph = coords
al, be = functions
metric = diag(
    -exp(2 * al(r)),
    exp(2 * be(r)),
    exp(2 * be(r)) * r ** 2,
    exp(2 * be(r)) * r ** 2 * sin(th) ** 2,
)



"""
Name: Minkowski
Coordinates: Spherical
Symmetry: Maximal
"""
from sympy import diag, sin, symbols

coords = symbols("t r theta phi", real=True)
variables = ()
functions = ()
t, r, th, ph = coords
metric = diag(-1, 1, r ** 2, r ** 2 * sin(th) ** 2)



"""
Name: Einstein
References: Stephani (10.23a) p122
Coordinates: Polar
Symmetry: Static
"""
from sympy import diag, sin, symbols

coords = symbols("t chi theta phi", real=True)
variables = symbols("Lambda", constant=True)
functions = ()
t, ch, th, ph = coords
La = variables
metric = diag(-1 / La, 1 / La, sin(ch) ** 2 / La, sin(ch) ** 2 * sin(th) ** 2 / La)



from riccipy.metric import *
from riccipy.tensor import *
from sympy import Array, diag, eye, simplify, sin, symbols
from sympy.tensor.tensor import TensExpr, TensMul


def _generate_simple():
    coords = symbols("x y z", real=True)
    metric = Metric("metric", coords, eye(3))
    return (coords, metric)


def _generate_schwarzschild():
    coords = symbols("t r theta phi", real=True)
    t, r, th, ph = coords
    schw = diag(1 - 1 / r, -1 / (1 - 1 / r), -(r ** 2), -(r ** 2) * sin(th) ** 2)
    g = Metric("g", coords, schw)
    mu, nu = indices("mu nu", g)
    return (coords, t, r, th, ph, schw, g, mu, nu)


def _generate_minkowski():
    coords = symbols("t x y z", real=True)
    t, x, y, z = coords
    mink = diag(1, -1, -1, -1)
    eta = Metric("eta", coords, mink)
    mu, nu = indices("mu nu", eta)
    return (coords, t, x, y, z, mink, eta, mu, nu)


def test_Index():
    (coords, metric) = _generate_simple()
    mu = Index("mu", metric)
    assert isinstance(mu, TensorIndex)
    assert mu.is_up
    assert not (-mu).is_up


def test_indices():
    (coords, metric) = _generate_simple()
    idxs = indices("i0:4", metric)
    assert len(idxs) == 4
    assert all([isinstance(i, Index) for i in idxs])


def test_Tensor():
    from sympy.tensor.tensor import TensorSymmetry

    (coords, metric) = _generate_simple()
    T = Tensor("T", coords, metric)
    assert isinstance(T, TensorHead)
    assert T.as_array() == Array(coords)
    assert T.covar == (1,)
    assert T.symmetry == TensorSymmetry.fully_symmetric(1)


def test_Tensor_comm():
    (coords, metric) = _generate_simple()
    T = Tensor("T", coords, metric)
    assert T.commutes_with(T) == 0
    assert T.commutes_with(metric) == 0
    assert T.commutes_with(metric.partial) is None


def test_Tensor_simplify():
    (coords, metric) = _generate_simple()
    x, y, z = coords
    expr = -((1 - x) ** 2) / (x - 1)
    expr_simplified = 1 - x
    T = Tensor("T", 3 * [expr], metric)
    assert str(T.simplify()[0]) == str(expr_simplified)


def test_Tensor_covariance_transform():
    (coords, t, r, th, ph, schw, g, mu, nu) = _generate_schwarzschild()
    E, p1, p2, p3 = symbols("E p_1:4", positive=True)
    p = Tensor("p", [E, p1, p2, p3], g)
    res = p.covariance_transform(-mu)
    expect = Array(
        [E * (1 - 1 / r), -p1 / (1 - 1 / r), -p2 * r ** 2, -p3 * r ** 2 * sin(th) ** 2]
    )
    assert res[0].equals(expect[0])
    assert res[1].equals(expect[1])
    assert res[2].equals(expect[2])
    assert res[3].equals(expect[3])


def test_AbstractTensor():
    (coords, metric) = _generate_simple()
    T = Tensor("T", coords, metric)
    assert isinstance(T, AbstractTensor)
    assert isinstance(T.as_array(), Array)
    assert T.is_Tensor
    assert not T.is_Metric
    assert not T.is_Spacetime


def test_IndexedTensor():
    (coords, metric) = _generate_simple()
    T = Tensor("T", coords, metric)
    mu = Index("mu", metric)
    assert isinstance(T(mu), IndexedTensor)


def test_expand_array():
    (coords, t, r, th, ph, mink, eta, mu, nu) = _generate_minkowski()
    E1, E2, E3, B1, B2, B3 = symbols("E_1:4 B_1:4", real=True)
    matrix = [[0, -E1, -E2, -E3], [E1, 0, -B3, B2], [E2, B3, 0, -B1], [E3, -B2, B1, 0]]
    F = Tensor("F", matrix, eta, symmetry=(2,))
    res1 = expand_array(eta(mu, nu) * eta(-mu, -nu))
    res2 = expand_array(eta(mu, -mu))
    assert res1 == res2 == 4
    assert expand_array(F(mu, -mu)) == 0
    assert (
        expand_array(F(mu, nu) * F(-mu, -nu))
        == 2 * B1 ** 2
        + 2 * B2 ** 2
        + 2 * B3 ** 2
        - 2 * E1 ** 2
        - 2 * E2 ** 2
        - 2 * E3 ** 2
    )
    (coords, t, r, th, ph, schw, g, mu, nu) = _generate_schwarzschild()
    x = Tensor("x", [t, r, th, ph], g)
    res = expand_array(g(mu, nu))
    assert schw.inv().equals(res)
    res1 = expand_array(g(mu, nu) * g(-mu, -nu))
    res2 = expand_array(g(-mu, -nu) * g(mu, nu))
    res3 = expand_array(g(-mu, mu))
    assert res1.equals(res2)
    assert res2.equals(res3)
    assert simplify(res1) == 4
    res1 = expand_array(g(-mu, -nu) * x(nu))
    res2 = expand_array(x(-mu))
    res3 = x.covariance_transform(-mu)
    assert res1 == res2 == res3
    res1 = expand_array(g(mu, nu) * x(-nu))
    res2 = expand_array(x(mu))
    assert res1.applyfunc(simplify) == res2
    res = expand_array(x(mu) * x(-mu))
    assert (
        res
        == t ** 2 * (1 - 1 / r)
        - r ** 2 / (1 - 1 / r)
        - th ** 2 * r ** 2
        - ph ** 2 * r ** 2 * sin(th) ** 2
    )


def test_expand_tensor():
    (coords, t, r, th, ph, schw, g, mu, nu) = _generate_schwarzschild()
    res = expand_tensor("t", g(-mu, -nu) * g(mu, nu), g)
    assert not isinstance(res, TensExpr)
    res = expand_tensor("T", g(mu, nu), g)
    assert isinstance(res, Tensor)
    assert schw.inv().equals(res.as_array())



"""
Name: LRS Stiff Perfect Fluid
References: Stephani (12.11) p146
Notes: Admits G4 on S3
"""
from sympy import diag, symbols

coords = symbols("t x y z", real=True)
variables = symbols("k", constant=True)
functions = ()
t, x, y, z = coords
k = variables
metric = diag(-1, t ** (2 / k), t ** (1 - 1 / k), t ** (1 - 1 / k))



"""
Name: Vaidya
References: Stephani (13.20) p158
Coordinates: Eddington-Finkelstein
Notes: Outgoing Coordinates
"""
from sympy import Function, diag, sin, symbols

coords = symbols("r v theta phi", real=True)
variables = ()
functions = symbols("M", cls=Function)
r, v, th, ph = coords
M = functions
metric = diag(0, -(1 - 2 * M(v) / r), r ** 2, r ** 2 * sin(th) ** 2)
metric[0, 1] = metric[1, 0] = -1



"""
Name: de Sitter
References: Hawking and Ellis p125
Coordinates: Cartesian
Symmetry: Maximal
"""
from sympy import diag, exp, symbols

coords = symbols("t x y z", real=True)
variables = symbols("alpha", constant=True)
functions = ()
t, x, y, z = coords
al = variables
expr = exp(2 * t / al)
metric = diag(-1, expr, expr, expr)



"""
Name: Bayin Perfect Fluid
References: Bayin, Phys. Rev. D, v18, p2745-2751, (1978)
Symmetry: Spherical
Coordinates: Spherical
"""
from sympy import Rational, asin, exp, sin, symbols, zeros

coords = symbols("t r theta phi", real=True)
variables = symbols("w_0 C_0 C_1", constant=True)
functions = ()
t, r, th, ph = coords
w0, C0, C1 = variables
metric = zeros(4)
metric[0, 0] = -(C0 ** 2) * exp(
    -asin((-2 * r ** 2 + C1) / (C1 ** 2 + 4 * w0 ** 2) ** (Rational(1, 2)))
)
metric[1, 1] = w0 ** 2 / (-(r ** 4) + C1 * r ** 2 + w0 ** 2)
metric[2, 2] = r ** 2
metric[3, 3] = r ** 2 * sin(th) ** 2



from collections import namedtuple

from sympy import Array, Pow, Rational, S, ones, tensorproduct, zeros
from sympy.tensor.tensor import TensorIndexType

from .partial import PartialDerivative, CovariantHead
from .tensor import AbstractTensor, Tensor, expand_array, indices


class Metric(AbstractTensor, TensorIndexType):
    """
    Class representing a tensor that raises and lowers indices.
    """

    # This object allows for having the metric be represented by the
    # same symbol as the tensor it associates with.
    _MetricId = namedtuple("MetricId", ["name", "antisym"])

    is_Metric = True
    _christoffel = None
    _riemann = None
    _ricci_tensor = None
    _ricci_scalar = None
    _weyl = None
    _einstein = None

    def __new__(cls, symbol, coords, matrix, **kwargs):
        """
        Create a new Metric object.

        Parameters
        ----------
        symbol : str
            Name of the tensor and the symbol to denote it by when printed.
        coords : iterable
            List of ~sympy.Symbol objects to denote the coordinates by which
            derivatives are taken with respect to.
        matrix : (list, tuple, ~sympy.Matrix, ~sympy.Array)
            Matrix representation of the tensor to be used in substitution.
            Can be of any type that is acceptable by ~sympy.Array.

        Examples
        --------
        >>> from sympy import diag, symbols
        >>> from riccipy import Metric, indices, expand_array
        >>> t, x, y, z = symbols('t x y z')
        >>> eta = Metric('eta', [t, x, y, z], diag(1, -1, -1, -1))
        >>> mu, nu = indices('mu nu', eta)
        >>> expr = eta(mu, nu) * eta(-mu, -nu)
        >>> expand_array(expr)
        4
        """
        array = Array(matrix)
        if array.rank() != 2 or array.shape[0] != array.shape[1]:
            raise ValueError(
                "matrix must be square, received matrix of shape {}".format(array.shape)
            )
        obj = TensorIndexType.__new__(
            cls,
            symbol,
            metric_name=cls._MetricId(symbol, False),
            dim=array.shape[0],
            dummy_name=symbol,
            **kwargs
        )
        obj = AbstractTensor.__new__(cls, obj, array)
        obj._metric = Tensor(obj.name, array, obj, covar=(-1, -1))
        obj.coords = tuple(coords)
        obj._repl[obj] = array
        return obj

    def __getattr__(self, attr):
        if hasattr(self.metric, attr):
            return getattr(self.metric, attr)
        return TensorIndexType.__getattribute__(self, attr)

    def __call__(self, *args):
        return self.metric(*args)

    def __getitem__(self, keys):
        return self.metric.__getitem__(keys)

    def subs(self, sub_dict):
        self.metric.subs(sub_dict)
        self._array = self._array.subs(sub_dict)
        self._repl[self] = self._array
        self._christoffel = None
        self._riemann = None
        self._ricci_tensor = None
        self._ricci_scalar = None
        self._weyl = None
        self._einstein = None

    def density(self, weight=S.One):
        return Pow(abs(self.determinant), Rational(weight, 2))

    @property
    def metric(self):
        return self._metric

    @property
    def determinant(self):
        return self.as_matrix().det()

    @property
    def partial(self):
        return PartialDerivative(self)

    @property
    def covariant(self):
        return CovariantHead(self)

    @property
    def christoffel(self):
        r"""
        Returns the Christoffel symbols using the formula:
        \Gamma^\sigma_{\mu\nu} =
        \frac{1}{2} g^{\sigma\rho} (\partial_\mu g_{\nu\rho} + \partial_\nu g_{\rho\mu} - \partial_\rho g_{\mu\nu})
        """
        if self._christoffel is None:
            mu, nu, si, rh = indices("mu nu sigma rho", self)
            d = self.partial
            g = self.metric
            gamma = (
                Rational(1, 2)
                * g(si, rh)
                * (d(-mu) * g(-nu, -rh) + d(-nu) * g(-rh, -mu) - d(-rh) * g(-mu, -nu))
            )
            syms = expand_array(gamma, [si, -mu, -nu])
            self._christoffel = Tensor("\u0393", syms, self, covar=(1, -1, -1))
        return self._christoffel

    @property
    def riemann(self):
        r"""
        Returns the Riemann curvature tensor using the formula:
        R^\rho_{\sigma\mu\nu} =
        \partial_\mu \Gamma^\rho_{\nu\sigma} - \partial_\nu \Gamma^\rho_{\mu\sigma}
        + \Gamma^\rho_{\mu\lambda} \Gamma^\lambda_{\nu\sigma} - \Gamma^\rho_{\nu\lambda} \Gamma^\lambda_{\mu\sigma}
        """
        if self._riemann is None:
            mu, nu, si, rh, la = indices("mu nu sigma rho lambda", self)
            d = self.partial
            gamma = self.christoffel
            R = (
                d(-mu) * gamma(rh, -nu, -si)
                - d(-nu) * gamma(rh, -mu, -si)
                + gamma(rh, -mu, -la) * gamma(la, -nu, -si)
                - gamma(rh, -nu, -la) * gamma(la, -mu, -si)
            )
            res = expand_array(R, [rh, -si, -mu, -nu])
            self._riemann = Tensor(
                "R", res, self, symmetry=(2, 2), covar=(1, -1, -1, -1)
            )
        return self._riemann

    @property
    def ricci_tensor(self):
        r"""
        Returns the Ricci tensor using the formula:
        R_{\mu\nu} = R^\sigma_{\mu\sigma\nu}
        """
        if self._ricci_tensor is None:
            mu, nu, si = indices("mu nu sigma", self)
            R = self.riemann
            res = expand_array(R(si, -mu, -si, -nu), [-mu, -nu])
            self._ricci_tensor = Tensor("R", res, self, covar=(-1, -1))
        return self._ricci_tensor

    @property
    def ricci_scalar(self):
        r"""
        Returns the Ricci scalar using the formula:
        R = R^\mu_\mu
        """
        if self._ricci_scalar is None:
            mu, nu = indices("mu nu", self)
            g = self.metric
            RR = self.ricci_tensor
            res = expand_array(g(-mu, -nu) * RR(mu, nu))
            self._ricci_scalar = res
        return self._ricci_scalar

    @property
    def weyl(self):
        r"""
        Returns the Weyl conformal tensor using the formula:
        C_{\rho\sigma\mu\nu} =
        R_{\rho\sigma\mu\nu} - \frac{2}{(n - 2)} (g_{\rho[\mu} R_{\nu]\sigma} - g_{\sigma[\mu} R_{\nu]\rho})
        + \frac{2}{(n - 1)(n - 2)} g_{\rho[\mu} g_{\nu]\sigma} R
        """
        if self._weyl is None:
            n = self.dim
            if n < 3:
                raise ValueError(
                    "the Weyl tensor is only defined in dimensions of 3 or more. {} is of dimension {}".format(
                        self, n
                    )
                )
            elif n == 3:
                res = tensorproduct(zeros(3, 3), zeros(3, 3))
                self._weyl = Tensor(
                    "C", res, self, symmetry=(2, 2), covar=(1, -1, -1, -1)
                )
                return self._weyl
            c1 = Rational(1, n - 2)
            c2 = Rational(1, (n - 2) * (n - 1))
            mu, nu, si, rh = indices("mu nu sigma rho", self)
            R = self.riemann
            RR = self.ricci_tensor
            RRR = self.ricci_scalar
            g = self.metric
            C = (
                R(rh, -si, -mu, -nu)
                - c1
                * (
                    g(rh, -mu) * RR(-nu, -si)
                    - g(rh, -nu) * RR(-mu, -si)
                    + g(-si, -nu) * RR(-mu, rh)
                    - g(-si, -mu) * RR(-nu, rh)
                )
                + c2 * (g(rh, -mu) * g(-nu, -si) - g(rh, -nu) * g(-mu, -si)) * RRR
            )
            res = expand_array(C, [rh, -si, -mu, -nu])
            self._weyl = Tensor("C", res, self, symmetry=(2, 2), covar=(1, -1, -1, -1))
        return self._weyl

    @property
    def einstein(self):
        r"""
        Returns the Einstein tensor using the formula:
        G_{\mu\nu} = R_{\mu\nu} - \frac{1}{2} R g_{\mu\nu}
        """
        if self._einstein is None:
            mu, nu = indices("mu nu", self)
            g = self.metric
            R = self.ricci_tensor
            RR = self.ricci_scalar
            res = expand_array(R(-mu, -nu) - Rational(1, 2) * RR * g(-mu, -nu))
            self._einstein = Tensor("G", res, self, covar=(-1, -1))
        return self._einstein


class SpacetimeMetric(Metric):
    """
    Class representing psuedo-Riemannian metrics.
    """

    is_Spacetime = True

    def __new__(cls, symbol, coords, matrix, timelike=True, **kwargs):
        obj = super().__new__(cls, symbol, coords, matrix, **kwargs)
        if obj.dim > 4:
            raise ValueError("metrics on spacetime must be at most 4-dimensional")
        obj.is_timelike = timelike
        obj.is_spacelike = not timelike
        return obj

    def reverse_signature(self):
        self._array *= -1
        self._replacement_dict = {self: self._array}
        self.is_timelike = not self.is_timelike
        self.is_spacelike = not self.is_spacelike
        return self.signature

    @property
    def signature(self):
        sign = -1 if self.is_timelike else 1
        sig = sign * ones(1, self.dim)
        sig[0] *= -1
        return tuple(sig)


def load_metric(symbol, name, coords=None, notes=None, timelike=False):
    from .metrics import data

    spacetime = data(name, coords=coords, notes=notes)
    if isinstance(spacetime, list):
        spacetime = spacetime[0]
    metric = SpacetimeMetric(
        symbol, spacetime["coords"], spacetime["metric"], timelike=False
    )
    if timelike:
        metric.reverse_signature()
    variables = spacetime["variables"]
    functions = spacetime["functions"]
    return (metric, variables, functions)



"""
Name: Kerr
Coordinates: Eddington Finkelstein
Symmetry: Axial
Notes: Outgoing Coordinates
"""
from sympy import cos, sin, symbols, zeros

coords = symbols("u r theta phi", real=True)
variables = symbols("a M", constant=True)
functions = ()
u, r, th, ph = coords
a, M = variables
expr = r ** 2 + a ** 2 * cos(th) ** 2
metric = zeros(4)
metric[0, 0] = -(a ** 2 - sin(th) ** 2 * a ** 2 + r ** 2 - 2 * M * r) / expr
metric[2, 2] = expr
metric[3, 3] = (
    sin(th) ** 2
    * (
        2 * a ** 2 * M * r * sin(th) ** 2
        + a ** 4
        - a ** 4 * sin(th) ** 2
        + 2 * r ** 2 * a ** 2
        - a ** 2 * sin(th) ** 2 * r ** 2
        + r ** 4
    )
    / expr
)
metric[0, 1] = metric[1, 0] = -1
metric[0, 3] = metric[3, 0] = -2 * a * sin(th) ** 2 * M * r / expr
metric[1, 3] = metric[3, 1] = a * sin(th) ** 2



"""
Name: Buchdahl-Land Perfect Fluid
References:
    - Tolman, Phys. Rev., v55, p363-373, (1939)
    - Buchdahl et al., J. Austr. Math. Soc., v8, p6-16, (1968)
    - Ibanez et al., J. Math. Phys., v23, p1363-1364, (1982)
    - Stephani (11.16) p128
Symmetry: Spherical
Coordinates: Spherical
"""
from sympy import Rational, diag, sin, symbols

coords = symbols("t r theta phi", real=True)
variables = ()
functions = ()
t, r, th, ph = coords
metric = diag(
    -Rational(2, 3) * r ** 2,
    1 / (Rational(1, 2) - r ** 2 / 3),
    r ** 2,
    r ** 2 * sin(th) ** 2,
)



from functools import partial
from itertools import starmap
from types import FunctionType

import numpy as np

from sympy import Array, lambdify, flatten
from sympy.tensor.tensor import TensExpr

from .tensor import expand_array


class NumericalArray(object):
    """
    Class for representing an array of symbolic expressions as lambdified functions.

    Calling instances of this class will result in the evaluation of a lambda function
    that returns a numerically valued array representing the results of the individual
    component expressions.

    Furthermore, instances may also be used to access specific components in a manner
    identical to accessing the components of an numpy array.
    """

    def __init__(self, args, array, **kwargs):
        """
        Create a new NumericalArray.

        Parameters
        ----------
        args : (list, tuple)
            Iterable of ~sympy.Symbol objects that specifcy the order of the arguments
            for the lambda functions that will be generated.
        array : (list, tuple, ~sympy.Matrix, ~sympy.Array)
            The array of symbolic expressions to be lambdified.
        """
        array = Array(array)
        self._vars = args
        self._objstr = str(array)
        self._array = array
        self._modules = kwargs.pop("modules", None)
        self._lambda = lambdify(self._vars, self._array, modules=self._modules)
        self._generator = self._build_generator()

    def _build_generator(self):
        shape = self._array.shape
        elements = flatten(self._array)
        ndarr = np.ndarray((len(elements),), dtype=FunctionType)
        lfunc = partial(lambdify, modules=self._modules)
        generator = starmap(lfunc, [(self._vars, elem) for elem in elements])
        ndarr.put(range(len(elements)), list(generator))
        return ndarr.reshape(shape)

    def __getitem__(self, key):
        return self._generator.__getitem__(key)

    def __call__(self, *args):
        return np.asarray(self._lambda(*args))

    def __getattr__(self, attr):
        if hasattr(self._array, attr):
            return getattr(self._array, attr)
        raise AttributeError("%s has no attribute: %s", self.__class__.__name__, attr)

    def __repr__(self):
        return "NumericalArray(%s)" % self._objstr

    def __str__(self):
        return self.__repr__()


def lambdify_tensor(args, expr, idxs=None, **kwargs):
    """
    Generate a numerical array representation for the result of a tensor expression.

    When the tensor expression passed results in an instance of ~sympy.Array, a
    NumericalArray object is returned.
    A NumericalArray can be used to access lambda functions for either the array
    as a whole or for individual components. See the example below.

    Parameters
    ----------
    args : (list, tuple)
        Iterable of ~sympy.Symbol objects that specifcy the order of the arguments
        for the lambda functions that will be generated.
    expr : (~sympy.tensor.tensor.TensExpr, ~sympy.Expr)
        The tensor expression to be lambdified. If this is an ordinary expression
        this function defaults to ~sympy.lambdify.
    idxs : (list, tuple)
        Iterable of Index objects to specify the indices for the result of ``expr``.

    Examples
    --------
    >>> from sympy import diag, sin, symbols
    >>> from riccipy import Metric, lambdify_tensor, indices
    >>> t, r, th, ph = symbols('t r theta phi', real=True)
    >>> schw = diag(1-1/r, -1/(1-1/r), -r**2, -r**2*sin(th)**2)
    >>> g = Metric('g', (t, r, th, ph), schw)
    >>> mu, nu = indices('mu nu', g)
    >>> narr = lambdify_tensor((r, th), g(-mu,-nu))
    >>> narr(2, 0)
    array([[ 0.5,  0. ,  0. ,  0. ],
           [ 0. , -2. ,  0. ,  0. ],
           [ 0. ,  0. , -4. ,  0. ],
           [ 0. ,  0. ,  0. , -0. ]])
    >>> narr[0,0]
    <function _lambdifygenerated at 0x...>
    >>> narr[0,0](2, 0)
    0.5
    """
    if isinstance(expr, TensExpr):
        expr = expand_array(expr, idxs)
    if isinstance(expr, Array):
        return NumericalArray(args, expr, **kwargs)
    return lambdify(args, expr, **kwargs)



from collections import deque
from sympy import Basic, Expr, Mul, S, Symbol, diff
from sympy.core.decorators import call_highest_priority
from sympy.tensor.tensor import (
    Tensor as SympyTensor,
    TensorHead,
    TensorSymmetry,
    TensMul,
)

from .tensor import Tensor, Index, IndexedTensor


class DiffOperator(Expr):
    _op_priority = 12
    is_commutative = False

    def __new__(cls, *args, left=S.One):
        if left == S.Zero:
            return S.Zero
        obj = Basic.__new__(cls, *args)
        obj._left = left
        return obj

    def __repr__(self):
        syms = [s.name for s in self.args]
        name = "\u2202(%s)" % ",".join(syms)
        return str(Mul(self._left, Symbol(name, commutative=False)))

    def __str__(self):
        return self.__repr__()

    @call_highest_priority("__rmul__")
    def __mul__(self, other):
        if isinstance(other, DiffOperator):
            args = self.args + other.args
            return DiffOperator(*args)
        return Mul(self._left, diff(other, *self.args)).doit()

    @call_highest_priority("__mul__")
    def __rmul__(self, other):
        return DiffOperator(*self.args, left=Mul(other, self._left))


class PartialDerivative(Tensor):
    is_TensorDerivative = True

    def __new__(cls, metric, **kwargs):
        basis = list(map(DiffOperator, metric.coords))
        return super().__new__(
            cls, "\u2202", basis, metric, comm="partial", covar=(-1,)
        )

    def __call__(self, idx):
        if idx.is_up:
            raise ValueError("partial derivatives cannot have a contravariant index")
        return IndexedPartial(self, [idx])

    def __repr__(self):
        return self._print()

    def __str__(self):
        return self.__repr__()


class IndexedPartial(IndexedTensor):
    _op_priority = 13

    @call_highest_priority("__rmul__")
    def __mul__(self, other):
        if not isinstance(other, TensMul):
            return SympyTensor.__mul__(self, other)
        expr = S.Zero
        terms = deque(other.args)
        for _ in range(len(terms)):
            terms.rotate(-1)
            leib, *niz = terms
            expr += TensMul(*niz, self * leib)
        return expr.doit()


class CovariantHead(TensorHead):
    is_TensorDerivative = True

    def __new__(cls, metric, **kwargs):
        sym = TensorSymmetry.fully_symmetric(1)
        obj = TensorHead.__new__(cls, "\u2207", [metric], sym, comm="partial")
        return obj

    def __call__(self, idx):
        return CovariantDerivative(self, [idx])


class CovariantDerivative(SympyTensor):
    _op_priority = 13
    is_TensorDerivative = True

    def __new__(cls, head, idx, left=S.One, **kwargs):
        obj = super().__new__(cls, head, idx, **kwargs)
        obj._head = head
        obj._idx = idx[0]
        obj._left = left
        return obj

    @call_highest_priority("__rmul__")
    def __mul__(self, other):
        metric = self._head.index_types[0]
        dum0 = Index(self._idx.name + "_0", metric)
        dum1 = Index(self._idx.name + "_1", metric, is_up=False)
        partial = metric.partial
        Gamma = metric.christoffel
        free_idxs = other.get_free_indices()
        coidx = self._idx
        if self._idx.is_up:
            coidx = dum1
        expr = partial(coidx) * other
        for idx in free_idxs:
            if idx.is_up:
                expr += Gamma(idx, coidx, -dum0) * other.substitute_indices((idx, dum0))
            else:
                expr -= Gamma(dum0, coidx, idx) * other.substitute_indices((idx, -dum0))
        left = self._left
        if self._idx.is_up:
            left *= metric(-coidx, self._idx)
        return TensMul(left, expr).doit()

    @call_highest_priority("__mul__")
    def __rmul__(self, other):
        left = TensMul(other, self._left).doit()
        return CovariantDerivative(self._head, [self._idx], left=left)

    def __repr__(self):
        tensor = SympyTensor(self._head, [self._idx])
        # TODO: Add printing support.
        return str(TensMul(self._left, tensor).doit())

    def __str__(self):
        return self.__repr__()



"""
Name: Durgapal
References: Durgapal, J. Phys. A, v15, p2637-2644, (1982)
Coordinates: Spherical
Symmetry:
    - Spherical
    - Static
Notes: n = 4
"""
from sympy import Rational, sin, symbols, zeros

coords = symbols("t r theta phi", real=True)
variables = symbols("A C K", constant=True)
functions = ()
t, r, th, ph = coords
A, C, K = variables
metric = zeros(4)
metric[0, 0] = -(A ** 2) * (1 + C * r ** 2) ** 4
metric[1, 1] = (
    7
    * (1 + C * r ** 2) ** 2
    * (1 + 5 * C * r ** 2) ** Rational(2, 5)
    / (
        (7 - 10 * C * r ** 2 - C ** 2 * r ** 4) * (1 + 5 * C * r ** 2) ** Rational(2, 5)
        + 7 * K * C * r ** 2
    )
)
metric[2, 2] = r ** 2
metric[3, 3] = r ** 2 * sin(th) ** 2



"""
Name: Godfrey
References:
    - Godfrey, Gen. Rel. Grav., v3, p3, (1972)
    - McIntosh, Gen. Rel. Grav., v7, p199-213, (1976)
Coordinates: Cylindrical
Notes:
    - Nontrivial Homothety
    - Not Hypersurface Orthogonal
    - Null Homothetic Bivector
"""
from sympy import diag, exp, symbols

coords = symbols("t r phi z", real=True)
variables = symbols("a C")
functions = ()
t, r, ph, z = coords
a, C = variables
expr = r ** (2 * a * (a - 1)) * exp(2 * (2 * a * z - z - r ** 2 / 2 + C))
metric = diag(-(r ** (2 * a)) * exp(2 * z), expr, r ** (2 * (1 - a)) / exp(2 * z), expr)



"""
Name: Gott Interior Cosmic String
References: Gott, Astrophys. J., v288, p422-427, (1985)
Coordinates: Cylindrical
"""
from sympy import diag, sin, symbols

coords = symbols("t r phi z", real=True)
variables = symbols("r_0", constant=True)
functions = ()
t, r, ph, z = coords
r0 = variables
metric = diag(-1, 1, r0 ** 2 * sin(r / r0) ** 2, 1)



"""
Name: Tolman
References: Tolman, Phys. Rev., v55, p363-373, (1939)
Coordinates: Spherical
Notes: Type VI
"""
from sympy import diag, sin, symbols

coords = symbols("t r theta phi", real=True)
variables = symbols("A B n", constant=True)
functions = ()
t, r, th, ph = coords
A, B, n = variables
metric = diag(
    -((A * r ** (1 - n) - B * r ** (1 + n)) ** 2),
    2 - n ** 2,
    r ** 2,
    r ** 2 * sin(th) ** 2,
)



"""
Name: de Sitter
References: Hawking and Ellis p125
Coordinates: Spherical
Symmetry: Maximal
Notes: Cosmological Constant
"""
from sympy import diag, sin, symbols

coords = symbols("t r theta phi", real=True)
variables = symbols("Lambda", constant=True)
functions = ()
t, r, th, ph = coords
La = variables
expr = 1 - La * r ** 2 / 3
metric = diag(-expr, 1 / expr, r ** 2, r ** 2 * sin(th) ** 2)



"""
RicciPy
=======

Python library for tensor manipulation in General Relativity
"""

from .metric import Metric, SpacetimeMetric, load_metric
from .numerical import lambdify_tensor
from .tensor import Index, Tensor, expand_array, expand_tensor, indices

__all__ = [
    "Metric",
    "SpacetimeMetric",
    "load_metric",
    "lambdify_tensor",
    "Index",
    "Tensor",
    "expand_array",
    "expand_tensor",
    "indices",
]



"""
Name: Melvin Magnetic Universe
References:
    - Bonnor, Prog. Roy. Soc. Lond., vA67, p225, (1954)
    - Melvin, Phys. Lett., v8, p65, (1964)
    - Stephani (20.10) p222
Coordinates: Cylindrical
"""
from sympy import diag, symbols

coords = symbols("t rho phi z", real=True)
variables = symbols("B_0", constant=True)
functions = ()
t, rh, ph, z = coords
B0 = variables
expr = (1 + B0 ** 2 * rh ** 2 / 4) ** 2
metric = diag(-expr, expr, rh ** 2 / expr, expr)



"""
Name: Lewis Papapetrou
References: Ernst, Phys. Rev., v167, p1175, (1968)
Coordinates: Cartesian
"""
from sympy import Function, Rational, exp, symbols, zeros

coords = symbols("t x y z", real=True)
variables = ()
functions = symbols("k r s w", cls=Function)
t, x, y, z = coords
k, r, s, w = functions
metric = zeros(4)
metric[0, 0] = -exp(2 * s(x, y))
metric[3, 3] = (exp(-s(x, y)) * r(x, y) - w(x, y) * exp(s(x, y))) * (
    exp(-s(x, y)) * r(x, y) + w(x, y) * exp(s(x, y))
)
metric[0, 3] = metric[3, 0] = -w(x, y) * exp(2 * s(x, y))
metric[1, 2] = metric[2, 1] = Rational(1, 2) * exp(2 * k(x, y) - 2 * s(x, y))



"""
Name: Szekeres Stiff Perfect Fluid
References: Szekeres, Commun. Math. Phys., v41, p55, (1975)
Coordinates: Cartesian
"""
from sympy import cos, diag, symbols

coords = symbols("t x y z", real=True)
variables = ()
functions = ()
t, x, y, z = coords
metric = diag(
    -(x ** 6), x ** 4 * cos(2 * t) ** 2, x ** 2 / cos(2 * t), x ** 2 / cos(2 * t)
)



"""
Name: Dunn and Tupper Perfect Fluid
Coordinates: Cartesian
"""
from sympy import diag, exp, symbols

coords = symbols("t x y z", real=True)
variables = symbols("b", constant=True)
functions = ()
t, x, y, z = coords
b = variables
metric = diag(
    -1,
    4 * t ** 2 / (-b * (1 + b)),
    t ** (-2 * b) * exp(-4 * x),
    t ** (-2 * b) * exp(4 * x),
)
